"""Unemployees look for work.

They're the mechanical drive behind the system.
"""

import logging
from ams.protocol import STATUS_DONE, STATUS_UNHANDLED, STATUS_FAILED
from ams.utils import endless_range

class BaseUnemployee(object):
    logger = logging.getLogger("ams.unemployee")

    def __init__(self, make_context, next_task):
        self.make_context = make_context
        self.next_task = next_task

    def __str__(self):
        spa = ["<%s " % type(self).__name__]
        if hasattr(self, "task_names_handled"):
            handles_str = ", ".join(self.task_names_handled)
            spa.append("for " + handles_str + " ")
        spa.append("using %s>" % self.next_task)
        return "".join(spa)

    def run_once(self, timeout=None):
        """Execute a singel task.

        This won't do anything if the task cannot be handled, and returns False
        in such a case.

        The task is burried if the execution fails, so that it may be rerun
        later.
        """
        task = self.next_task(timeout=timeout)
        if not self.handles_task(task):
            self.logger.warn("%r unable handle %r", self, task)
            task.set_status(STATUS_UNHANDLED)
            return False
        try:
            self.execute_task(task)
        except:
            self.logger.exception("%r execute_task %r", self, task)
            task.set_status(STATUS_FAILED)
        else:
            self.logger.debug("%r done %r", self, task)
            task.set_status(STATUS_DONE)
            return True

    def run(self, times=None):
        """Execute tasks, stopping after *times* tasks have been run. This
        doesn't necessarily mean that *times* tasks have been executed.

        If *times* is None, runs indefinitely.
        """
        if times:
            counter = xrange(times)
        else:
            counter = endless_range()
        for count in counter:
            self.logger.debug("looking for a job (#%d)", count)
            self.run_once()

    # Unemployee subclass API
    def handles_task(self, task):
        """Check whether or not the unemployee handles the task.

        This could be for any number of reasons, but will mostly be because the
        instance doesn't know how to handle the task type specified.

        Subclasses must override this and provide logics for determining this
        themselves.
        """
        raise NotImplementedError()

    def execute_task(self, task):
        """Execute a given task.

        This task should have been run through handles_task first, to see if
        it's feasible for execution. That is to say, handles_task shouldn't be
        run here.

        Subclasses must override this and provide logics for dispatching tasks
        to the correct task handler.
        """
        return self.handler_for_task(task).run_task(task)

    def handler_for_task(self, task):
        """Get the handling function for *task*.

        Only used by *execute_task*, and as such may not be relevant.

        The return value is expected to have a *run_task* attribute, which runs
        the actual task.

        Subclasses must override this or *execute_task* to provide logics for
        getting the handler for the correct task.
        """
        raise NotImplementedError()

class SimpleUnemployee(BaseUnemployee):
    """Executes jobs of a single type."""

    def __init__(self, make_context, next_task, handler):
        super(SimpleUnemployee, self).__init__(make_context, next_task)
        self.handler = handler

    @property
    def task_names_handled(self):
        return [self.handler.type_name]

    def handles_task(self, task):
        return task.type_name == self.handler.type_name

    def handler_for_task(self, task):
        return self.handler(self.make_context)

class FlexibleUnemployee(BaseUnemployee):
    """Executes jobs of multiple types."""

    def __init__(self, make_context, next_task, handlers):
        super(FlexibleUnemployee, self).__init__(make_context, next_task)
        self.handler_map = dict((h.type_name, h) for h in handlers)

    @property
    def task_names_handled(self):
        return self.handler_map.keys()

    def handles_task(self, task):
        return task.type_name in self.handler_map

    def handler_for_task(self, task):
        return self.handler_map[task.type_name](self.make_context)



"""Protocol-related functionality."""

try:
    import cPickle as pickle
except ImportError:
    import pickle

STATUS_DONE = "done"
STATUS_UNHANDLED = "unhandled"
STATUS_FAILED = "failed"
STATUSES = STATUS_DONE, STATUS_UNHANDLED, STATUS_FAILED

class BaseTask(object):
    """Tasks are the abstraction layer between the queue server's job
    representation and the generic interface used throughout AMS.

    See the package's docs for more info on the overall workings.
    """

    def __init__(self, type_name, args):
        """Initialize a new task with type *type_name*, and arguments *args*."""
        self.type_name = type_name
        self.args = args

    def __repr__(self):
        return "%s(%r, %r)" % (type(self).__name__, self.type_name, self.args)

    def __eq__(self, other):
        if hasattr(other, "type_name") and hasattr(other, "args"):
            return other.type_name == self.type_name and other.args == self.args
        return NotImplemented

    @staticmethod
    def decode(raw):
        """Decode raw job data *raw* into a two-tuple of (task type, args).

        The format for the raw data is JSON, with the outermost value being an
        object with at least a type member. The rest of the members are
        arguments. (The type member is removed from args in the return tuple.)
        """
        data = pickle.loads(raw)
        return (data.pop("type"), data)

    def encode(self):
        """Encode the instance into raw job data.

        See the decode method for the data format used.
        """
        data = self.args.copy()
        data["type"] = self.type_name
        return pickle.dumps(data)

    def set_status(self, status):
        """Set status of this task to *status*. The statuses may be one of
        STATUSES.

        If the status is STATUS_DONE, the queue job should be flagged as
        such, so it's not run again.

        If the status is STATUS_FAILED, the queue job should be somehow stored
        so it may be run at a later time.

        If the status is STATUS_UNHANDLED, the queue job should be reinserted
        into the queue server so another pass of next_task might get it again.
        """
        raise NotImplementedError()

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""Configuration system for AMS."""

class BaseConf(object):
    """Provides the basic interface for configuration objects."""

    @property
    def db_type(self):
        """Database type (currently only 'postgresql' is allowed.)

        Subclasses need to override this.
        """
        raise NotImplementedError()

    @property
    def db_info(self):
        """Database information in format of `(name, user, pswd, host, port)`.

        Subclasses need to override this.
        """
        raise NotImplementedError()

class DummyConf(BaseConf):
    """Dummy conf type which reads no configuration at all."""

    dsn_name = "dummy"

    @classmethod
    def from_dsn(cls, dsn):
        if dsn.geturl() != "dummy:":
            raise ValueError("dummy conf doesn't accept options: %r" % (dsn,))
        return cls()

# {{{ Django settings descriptors
# TODO This is essentially an attribute getter.
class DjangoSettingGetter(object):
    def __init__(self, name, settings=None):
        self.name = name

    def __get__(self, instance, owner):
        return getattr(instance.mod, self.name)

class DjangoSettingsGetter(object):
    def __init__(self, names, prefix=None):
        prefix = prefix or ""
        self.names = tuple(prefix + name for name in names)

    def __get__(self, instance, owner):
        return tuple(getattr(instance.mod, name) for name in self.names)
# }}}

class DjangoConf(BaseConf):
    dsn_name = "django"

    def __init__(self, mod, setup_environ=True):
        self.mod = mod
        if setup_environ:
            self.setup_environ(mod)

    def __str__(self):
        return "<%s using %s>" % (type(self).__name__, self.mod)

    @classmethod
    def from_dsn(cls, dsn):
        # Don't auto-setup the environ since we have a better name for the
        # module.
        modname = dsn.path
        mod = __import__(modname, fromlist=[""])
        self = cls(mod, setup_environ=False)
        self.setup_environ(mod, modname)
        return self

    db_type = DjangoSettingGetter("DATABASE_ENGINE")
    db_info = DjangoSettingsGetter(("NAME", "USER", "PASSWORD",
                                    "HOST", "PORT"), prefix="DATABASE_")

    def setup_environ(self, mod, modname=None):
        from django.core.management import setup_environ
        return setup_environ(mod, original_settings_path=modname)

class PyconfConf(BaseConf):
    dsn_name = "py"

    def __init__(self, *args, **kwds):
        raise TypeError("cannot make instances of %r" % (self.__class__,))

    @classmethod
    def from_dsn(cls, dsn):
        return pyconf.load(dsn.path)

try:
    import pyconf
except ImportError:
    del PyconfConf

all_conf_types = list(BaseConf.__subclasses__())



"""Misc. utilities."""

def endless_range(start=0, step=1):
    i = start
    while True:
        yield i
        i += step

try:
    from setproctitle import setproctitle
except ImportError:
    setproctitle = lambda t: NotImplemented

def mixined(cls, *mixin_clses):
    return type(cls.__name__ + "+Mixins", mixin_clses + (cls,), {})

import sys
import imp
from os import walk, path

def load_whole_package(package_modname):
    package = __import__(package_modname, fromlist=[""], level=0)
    suffixes = imp.get_suffixes()
    origin = path.dirname(package.__file__)
    for dirn, subdirns, fnames in walk(origin):
        subdirns[:] = filter(lambda dn: not dn.startswith("."), subdirns)
        curr_package = package_modname + dirn[len(origin):].replace(path.sep, ".")
        found_mods = {}
        # Phase one: collect modules to load in this directory.
        for fname in fnames:
            for suffix, mode, type_ in suffixes:
                if fname.endswith(suffix):
                    fpath = path.join(dirn, fname)
                    desc = suffix, mode, type_
                    found_mods[fname[:-len(suffix)]] = (fpath, desc)
        # Phase two: load modules in order (to get __init__ first and some
        # predictability.)
        for modname in sorted(found_mods):
            fpath, desc = found_mods[modname]
            if modname == "__init__":
                modname = curr_package
            else:
                modname = curr_package + "." + modname
            # Only load modules if they aren't already.
            if modname not in sys.modules:
                suffix, mode, type_ = desc
                fp = open(fpath, mode)
                try:
                    mod = imp.load_module(modname, fp, fpath, desc)
                finally:
                    fp.close()
                imp.acquire_lock()
                sys.modules[modname] = mod
                imp.release_lock()

# DSN Parsing
import urlparse

class DSNParseError(Exception): pass

def parse_dsn(dsn_url, scheme_map, default_scheme=None):
    """Parse *dsn_url* into a tuple of `(cls, opts)`."""
    dsn = urlparse.urlsplit(dsn_url)
    scheme = dsn.scheme or default_scheme
    if scheme not in scheme_map:
        raise DSNParseError("scheme must be one of %r, not %r" %
                            (scheme_map.keys(), dsn.scheme))
    return scheme_map[scheme], dsn



"""AMS -- takes care of handing out jobs to unemployees.

The fundamental concept in AMS is the task type. Each task is of a specific
task type, and each unemployee can handle any number of task types through the
use of task handlers.

The task handlers are simply callables which take at least a conf argument, and
any number of keyword arguments which are passed on as task data.

The job data in the queue system for AMS is serialized using JSON.
"""



import sys

import logging
from logging.handlers import WatchedFileHandler

log_format = "%(levelname)9s [%(asctime)s] %(name)s: %(message)s"

# TODO Do to ams.conf what we did to ams.platsbanken.
from ams.conf import all_conf_types
from ams.utils import DSNParseError, setproctitle, mixined, parse_dsn
from ams.platsbanken import get_all_platsbanker, parse_platsbank_dsn
from ams.unemployee import FlexibleUnemployee

default_platsbank_scheme = get_all_platsbanker()[0].dsn_name

def parse_conf_dsn(dsn, conf_types=all_conf_types):
    conf_type_map = dict((ct.dsn_name, ct) for ct in conf_types)
    conf_cls, opts = parse_dsn(dsn, conf_type_map)
    return conf_cls.from_dsn(opts)

import optparse

class SetProcessTitleMixin(object):
    def run_once(self, *a, **kw):
        hstr = ", ".join(self.task_names_handled)
        if len(hstr) > 50:
            hstr = hstr[:47] + "..."
        setproctitle("ams-unemployee: waiting (%s)" % hstr)
        return super(SetProcessTitleMixin, self).run_once(*a, **kw)

    def execute_task(self, task):
        setproctitle("ams-unemployee: executing (%s)" % (task.type_name,))
        return super(SetProcessTitleMixin, self).execute_task(task)

class UnemployeeApp(object):
    parser = optparse.OptionParser()
    parser.add_option("-p", "--platsbank", dest="platsbank_dsn", metavar="DSN",
                      help="Use DSN to connect to Platsbanken.")
    parser.add_option("-c", "--conf", dest="conf_dsn", metavar="CONF",
                      help="Use CONF for configuration.")
    parser.add_option("-v", "--verbose", dest="verbosity", action="count",
                      default=0, help="Increase logging verbosity.")
    parser.add_option("-l", "--log", dest="log_target", default="-",
                      help="Log to target file. Use - for stderr.")
    parser.add_option("-n", "--times", dest="times", metavar="N", type="int",
                      help="Run max N jobs (default: no limit).")

    logger = logging.getLogger("ams.unemployee.app")
    log_levels = logging.DEBUG, logging.INFO, logging.WARN, logging.ERROR
    log_format = "%(levelname)9s [%(asctime)s] %(name)s: %(message)s"
    log_date_format = None

    unemployee_cls = mixined(FlexibleUnemployee, SetProcessTitleMixin)

    def __init__(self, args):
        self.opts, self.args = self.parse_args(args)
        self.setup_logging(self.opts)
        self.conf = self.load_conf(self.opts)
        self.logger.info("using conf %s", self.conf)
        self.platsbank = self.setup_platsbanken(self.opts)
        self.logger.info("platsbank: %s", self.platsbank)
        self.unemployee = self.build_unemployee()

    parse_args = parser.parse_args

    def invalid_setup(self, reason):
        self.logger.error("invalid setup: %s", reason)
        sys.exit(1)

    # {{{ Logging
    def _log_handler(self, target, handler):
        """Set *handler* as a logging handler for *target*, and set some stuff."""
        handler.setFormatter(self.make_logging_formatter())
        target.addHandler(handler)

    def make_logging_formatter(self):
        return logging.Formatter(self.log_format, self.log_date_format)

    def setup_logging(self, opts):
        self.log_level = self.log_levels[max(0, 2 - opts.verbosity)]
        logging.root.setLevel(self.log_level)
        if opts.log_target != "-":
            self.setup_logging_file(self.log_level, opts.log_target)
        else:
            self.setup_logging_stderr(self.log_level)

    def setup_logging_file(self, level, fname, target=logging.root):
        handler = WatchedFileHandler(fname)
        handler.setLevel(level)
        self._log_handler(target, handler)

    def setup_logging_stderr(self, level, target=logging.root):
        handler = logging.StreamHandler()
        handler.setLevel(level)
        self._log_handler(target, handler)
    # }}}

    def load_conf(self, opts):
        dsn = self.get_conf_dsn(opts)
        try:
            conf = parse_conf_dsn(dsn)
        except DSNParseError, e:
            self.invalid_setup("dsn %r: %s" % (dsn, e))
        return conf

    def get_conf_dsn(self, opts):
        return opts.conf_dsn

    def setup_platsbanken(self, opts):
        dsn = self.get_platsbanken_dsn(opts)
        try:
            platsbank = parse_platsbank_dsn(dsn)
        except DSNParseError, e:
            self.invalid_setup("dsn %r: %s" % (dsn, e))
        return platsbank

    def get_platsbanken_dsn(self, opts):
        if opts.platsbank_dsn:
            return opts.platsbank_dsn
        elif hasattr(self.conf, "platsbank_dsn"):
            return self.conf.platsbank_dsn
        else:
            return default_platsbank_scheme + ":"

    def build_unemployee(self):
        hdlrs = self.load_task_handlers()
        self.platsbank.accept_task_types(hdlr.type_name for hdlr in hdlrs)
        return self.unemployee_cls(self.conf, self.platsbank.next_task, hdlrs)

    def run_unemployee(self, unemp=None):
        if unemp is None:
            unemp = self.unemployee
        self.logger.info("running %s %s times", unemp,
                         self.opts.times or "unlimited")
        unemp.run(self.opts.times)

    def load_task_handlers(self):
        specs = self.args
        if not specs:
            specs = list(getattr(self.conf, "task_handlers", []))
            if not specs:
                self.invalid_setup("no task handlers specs")
        rv = []
        for spec in specs:
            rv.extend(self.load_task_handler(spec))
        return rv

    def load_task_handler(self, spec):
        rv = []
        modname, attrs = spec.rsplit(".", 1)
        self.logger.info("searching module %s for %s", modname, attrs)
        mod = __import__(modname, fromlist=[""])
        self.logger.debug("found module %r", mod)
        # If the specification is *, load by the attribute handlers on the
        # module.
        if attrs == "*":
            if not hasattr(mod, "handlers"):
                raise ValueError("couldn't load specifier %s: module %s has no "
                                 "'handlers' attribute" % (spec, mod))
            self.logger.info("adding task handlers %r", mod.handlers)
            rv = mod.handlers
        else:
            # If we have a specification like {a,b,c}, split it out.
            if attrs.startswith("{") and attrs.endswith("}"):
                attnames = attrs[1:-1].split(",")
            else:
                attnames = (attrs,)
            # Then add the task types one by one.
            for attname in attnames:
                handler = getattr(mod, attname)
                self.logger.info("adding task handler %r", handler)
                rv.append(handler)
        return rv

    @classmethod
    def main(cls):
        self = cls(sys.argv[1:])
        self.run_unemployee()

if __name__ == "__main__":
    UnemployeeApp.main()



"""Platsbanken provides interfaces to AMS for both giving and getting jobs."""

from ams.protocol import BaseTask
from ams.utils import parse_dsn, load_whole_package

class Platsbank(object):
    def next_task(self, timeout=None):
        """Fetch the next task from the queue server, blocking for max
        *timeout* seconds.
        """
        raise NotImplementedError()

    def accept_task_types(self, type_names=None):
        """Accept tasks of type name in *type_names*. If *type_names* is None,
        all types are fair game.
        """
        raise NotImplementedError()

    def queue_task(self, priority=None, delay=None, ttr=None):
        """Queue this task on the queue server, setting priority to *priority*,
        and delay insertion to after *delay* seconds.

        Note that the job should appear after *delay* seconds, not that this
        method should block for *delay* seconds before inserting.

        If *ttr* is specified, it is the number of seconds a task may run.

        As above, if the queue server doesn't handle any or all of the
        features, a ValueError should be raised if that keyword argument is not
        None.

        This method is for subclasses to implement in the best possible way.
        """
        raise NotImplementedError()

    def queue(self, type_name, **args):
        """Queue a task of type *type_name* with arguments *args*.

        Really just calls self.queue_task with a task object constructed from
        the arguments given to this function.
        """
        pass_kws = ("priority", "delay", "ttr")
        kwds = dict((k, v) for (k, v) in args.items() if k in pass_kws)
        return self.queue_task(BaseTask(type_name, args), **kwds)

class DummyPlatsbank(Platsbank):
    dsn_name = "dummy"

    def next_task(self, timeout=None):
        raise RuntimeError("dummy platsbank can't get tasks")

    def queue_task(self, task, priority=None, delay=None, ttr=None):
        pass

    def accept_task_types(self, type_names=None):
        self.accept_type_names = type_names

    @classmethod
    def from_dsn(cls, dsn):
        return cls()

class RecordingDummyPlatsbank(DummyPlatsbank):
    dsn_name = "dummy+record"

    def __init__(self):
        self.queued = []

    def queue_task(self, task, priority=None, delay=None):
        # Funny way to copy the base task type.
        task = type(task)(task.type_name, task.args)
        task.priority = priority
        task.delay = delay
        self.queued.append(task)
        return task

_platsbanker_cache = []
def parse_platsbank_dsn(dsn, platsbanker=None, default=None):
    if platsbanker is None:
        platsbanker = get_all_platsbanker()
    platsbank_map = dict((pb.dsn_name, pb) for pb in platsbanker)
    platsbank_cls, opts = parse_dsn(dsn, platsbank_map, default)
    return platsbank_cls.from_dsn(opts)

def get_all_platsbanker():
    rv = _platsbanker_cache
    if not rv:
        load_whole_package("ams.platsbanken")
        # FIXME Don't use __subclasses__, only nests one level.
        def _walk_clses(cls):
            if hasattr(cls, "dsn_name") and hasattr(cls, "from_dsn"):
                rv.append(cls)
            for subcls in cls.__subclasses__():
                _walk_clses(subcls)
        _walk_clses(Platsbank)
    return rv



"""Beanstalk source for platsbanken."""

from __future__ import absolute_import

import logging
from ams.platsbanken import Platsbank
from ams.protocol import (BaseTask, STATUSES, STATUS_DONE,
                          STATUS_UNHANDLED, STATUS_FAILED)

# Let this fail later on if the module doesn't exist
try:
    import beanstalkc
except ImportError:
    beanstalkc = None

logger = logging.getLogger("ams.unemployee.platsbanken.beanstalk")

class BeanstalkTask(BaseTask):
    """Beanstalkd-specific task subclass."""

    def __init__(self, job, type_name, args):
        """Annotates the instance with a *job* attribute."""
        super(BeanstalkTask, self).__init__(type_name, args)
        self.job = job

    def __hash__(self):
        return hash(self.job.jid)

    @classmethod
    def from_job(cls, job):
        """Create a task from the beanstalkc job instance *job*."""
        type_name, args = cls.decode(job.body)
        return cls(job, type_name, args)

    def set_status(self, status):
        """Set job status of this task to *status*.
        
        STATUS_DONE means the job is deleted from the queue,
        STATUS_UNHANDLED means the job is simply released back into the queue,
        STATUS_FAILED means the job is burried.
        """
        if status not in STATUSES:
            raise ValueError("status %r unknown (should be one of %r)" %
                             (status, STATUSES))
        if status == STATUS_DONE:
            self.job.delete()
        # TODO Better workaround than giving a false priority in release and
        #      bury.
        # If you don't pass a priority to bury and release, beanstalkc will try
        # to look up the job's priority (why doesn't beanstalkd do this by
        # itself?) through decoding the YAML stats of the job.
        # Not a great idea, and top that off with the fact that we don't load
        # YAML.
        # What needs to be done is that beanstalkd has a sensible default
        # priority for burried and released jobs (i.e., the job's current
        # priority.)
        elif status == STATUS_UNHANDLED:
            self.job.release(priority=beanstalkc.DEFAULT_PRIORITY)
        elif status == STATUS_FAILED:
            self.job.bury(priority=beanstalkc.DEFAULT_PRIORITY)

class BeanstalkPlatsbank(Platsbank):
    """Platsbanken for beanstalkd."""

    dsn_name = "beanstalk"
    task_cls = BeanstalkTask

    def __init__(self, client):
        self.client = client

    def __str__(self):
        return "<%s with %s>" % (type(self).__name__, self.client)

    def next_task(self, timeout=None):
        """Get next task from beanstalkd."""
        job = self.client.reserve(timeout=timeout)
        if job:
            return self.task_cls.from_job(job)

    def queue_task(self, task, priority=None, delay=None, ttr=None):
        """Queue task *task* in beanstalkd and return updated task object."""
        if priority is None: priority = beanstalkc.DEFAULT_PRIORITY
        if delay is None: delay = 0
        if ttr is None: ttr = beanstalkc.DEFAULT_TTR
        body = task.encode()
        self.client.use(task.type_name)
        jid = self.client.put(body, priority=priority, delay=delay, ttr=ttr)
        job = beanstalkc.Job(self.client, jid, body, reserved=False)
        return self.task_cls.from_job(job)

    def accept_task_types(self, type_names=None):
        if type_names is None:
            return
        for type_name in type_names:
            self.client.watch(type_name)

    @classmethod
    def from_dsn(cls, dsn):
        """Create a platsbank connected to the beanstalkd at *dsn*."""
        if not dsn.path:
            host, port = beanstalkc.DEFAULT_HOST, beanstalkc.DEFAULT_PORT
        elif ":" in dsn.path:
            host, port = dsn.path.split(":")
            port = int(port)
        else:
            host, port = dsn.path, beanstalkc.DEFAULT_PORT
        logger.info("%s connecting to (%r, %r)", cls, host, port)
        conn = beanstalkc.Connection(host=host, port=port)
        return cls(conn)



class Handler(object):
    """Handler for tasks of a specific type."""

    def __init__(self, conf):
        self.conf = conf

    def run_task(self, task):
        return self(**dict((str(k), v) for (k, v) in task.args.iteritems()))

    def __call__(self, **kwds):
        raise NotImplementedError("%r doesn't implement any handling code "
                                  "(kwds: %r)" % (self, kwds))

class SimpleHandler(Handler):
    """Single-function handlers are subclasses of this class, with __call__ set
    to the handler function.
    """

def conf_passer(func):
    """Passes the conf instance of a handler as the first argument instead of
    self.
    """
    def inner(self, *a, **kw):
        return func(self.conf, *a, **kw)
    return inner

def simple_handler(name, bases=(SimpleHandler,)):
    def deco(func):
        clsname = func.__name__.replace("_", " ").title().replace(" ", "")
        attrs = {"__call__": conf_passer(func), "type_name": name}
        subcls = type(clsname, bases, attrs)
        return subcls
    return deco

# Here's some fine meta class trickery. RegisterHandlerType is a meta type
# which adds newly created classes of its type to a handlers registry.
class RegisterHandlerType(type):
    def __new__(cls, name, bases, attrs):
        rv = super(RegisterHandlerType, cls).__new__(cls, name, bases, attrs)
        if "_handlers" not in attrs:
            rv._handlers.append(rv)
        return rv

class HandlerRegistry(list):
    """Helper for registering handlers in a module."""

    def __init__(self, *args, **kwds):
        super(HandlerRegistry, self).__init__(*args, **kwds)
        # The aforementioned meta type is subclassed and "bound" to this
        # instance through the use of per-instance subclasses and a _handlers
        # class attribute.
        attrs = {"_handlers": self}
        cls_type = RegisterHandlerType
        self.Handler = cls_type("RHandler", (Handler,), attrs)
        self.SimpleHandler = cls_type("RSimpleHandler", (SimpleHandler,), attrs)

    def simple_handler(self, name):
        """Just like handlers.simple_handler, but registers the handler."""
        # Quite straight-forward, sets base class to our special-type class.
        return simple_handler(name, bases=(self.SimpleHandler,))



"""Simple task handlers like ping, error, etc."""

from ams.unemployee.handlers import HandlerRegistry
from time import sleep
import logging

def make_logger(subname=""):
    if subname:
        subname = "." + subname
    return logging.getLogger("ams.unemployee.handlers.simple" + subname)

logger = make_logger()
handlers = HandlerRegistry()

@handlers.simple_handler("ping")
def log_pings(conf, data):
    make_logger("ping").info("ping: %r", data)

@handlers.simple_handler("raise-error")
def raise_error(conf, value):
    raise TypeError("phony error (%r)" % (value,))

@handlers.simple_handler("sleep")
def do_sleep(conf, delay=5.0):
    sleep(delay)
